/*******************************************************************

 joinvect.c : Joins together VECT curve.
           See the man page for details.
	   
	   ************************************************************/

#include<joinvect.h>
#include<argtable2.h>

/**********************************************************************/

struct arg_lit *verbose;
struct arg_file *corefile;
struct arg_lit  *help;
struct arg_lit  *fullauto;

struct arg_end *end;
struct arg_end *helpend;

plCurve *core;
FILE         *infile_fptr,*outfile_fptr;

/****************************** Main procedure ********************************/
  
int main(int argc,char *argv[])
{
  int            infilenum,nerrors;
  char           outfile_name[1000],outfile_tail[1000];
 
  void *argtable[] = 
    {
     verbose = arg_lit0("v","verbose","print debugging information"),
     corefile  = arg_filen(NULL,NULL,"<file>",1,100000,"input files"),
     fullauto = arg_lit0("a","automatic-mode","decides on join order"),
     help = arg_lit0(NULL,"help","display help message"),
     end = arg_end(20)};
  
  void *helptable[] = {help,helpend = arg_end(20)};
  
  /* First, we parse the command-line arguments using argtable. */

  if (arg_nullcheck(argtable) != 0)
    printf("splitvect: Insufficient memory to allocate argument table.\n");

  nerrors = arg_parse(argc,argv,argtable);

  if (nerrors > 0) { /* The standard syntax didn't match-- try the help table */
    
    nerrors = arg_parse(argc,argv,helptable);
    
    if (nerrors > 0) {  /* The help table didn't match either-- the
                         first set of errors was probably more
                         helpful, so we display it. */

      arg_print_errors(stdout,end,"joinvect");
      exit(1);

    } else {  /* The help table matched, which means we 
		 asked for help or gave nothing */
  
      printf("joinvect concatenates VECT files\n"
           "usage: \n\n");
      arg_print_glossary(stdout, argtable," %-25s %s\n");
      exit(0);

    }
    
  }
  
  /* Now we have parsed the arguments and are ready to work. */
 
  plCurve **curves;
  curves = calloc(corefile->count,sizeof(plCurve *));
  int ncurves = 0;
 
  for(infilenum = 0;infilenum < corefile->count;infilenum++) {

    /* We begin by loading the current core link. */

     infile_fptr = fopen(corefile->filename[infilenum],"r");
  
     if (infile_fptr == NULL) {
    
       fprintf(stderr,"joinvect: Couldn't open file %s.\n",corefile->filename[infilenum]);
       continue;  /* Try the next file */
 
     }
  
     int plr_error_num;
     char plr_error_str[1024];

     curves[ncurves++] = plc_read(infile_fptr,
				  &plr_error_num,plr_error_str,sizeof(plr_error_str));
  
     if (plr_error_num > 0) {   /* This is the signal for an error. */
    
       fprintf(stderr,"joinvect: link reading error\n%s\n",plr_error_str);
       ncurves--;
       continue;  /* Try the next file */
    
     }
  
     if (curves[ncurves-1]->nc > 1) {

       fprintf(stderr,
	       "joinvect: Curve %s has more than one component. \n"
	       "          Use splitvect to separate components \n"
	       "          before joining.\n",corefile->filename[infilenum]);

       plc_free(curves[ncurves-1]);
       ncurves--;
       continue;
     }

     if (curves[ncurves-1]->cp[0].open == false) {

       fprintf(stderr,
	       "joinvect: Curve %s is already closed. Skipping it.\n",
	       corefile->filename[infilenum]);

       plc_free(curves[ncurves-1]);
       ncurves--;
       continue;
     }

     fclose(infile_fptr);

  }

  printf("joinvect: Loaded %d curves from VECT files.\n",ncurves);

  /* Now determine order. */

  int *order;
  order = calloc(ncurves,sizeof(int));
  int i;

  for(i=0;i<ncurves;i++) {

    order[i] = i+1;

  }

  order[ncurves-1]=0;

  if (fullauto->count > 0) { 

    printf("joinvect: Searching for optimal join order using travelling salesman algorithm...\n");
    tsp_order(ncurves,curves,order);

  }

  /* Now actually join the curves. */

  plCurve *joined;
  bool    *open;
  int     *cc;
  bool    *used;

  /* Here's the rub... we may have more than one component here. */

  used = calloc(ncurves,sizeof(bool));

  int nc=0,startc,thisc;
  int *nv;

  /* First, determine the number of loops and the size of each. */

  nv = calloc(ncurves,sizeof(int));
  open = calloc(ncurves,sizeof(bool));
  cc = calloc(ncurves,sizeof(int));

  for(startc=0;startc<ncurves;startc++) {

    if (!used[startc]) {

      used[startc] = true;
      nv[nc] += plc_num_verts(curves[startc]);
      
      for(thisc=order[startc];thisc != startc;thisc = order[thisc]) {
	
	nv[nc] += plc_num_verts(curves[thisc]);
	used[thisc] = true;

      }

      nc++;

    }

  }

  printf("\njoinvect: Joined curves have %d components.\n",nc);
  joined = plc_new(nc,nv,open,cc);

  /* Now run the same loop, adding the actual vertices to the curves. */

  int cv;  
  nc = 0;

  for(i=0;i<ncurves;i++) { used[i] = false; }

  for(startc=0;startc<ncurves;startc++) {
    
    if (!used[startc]) {

      printf("joinvect: Component loop %d-",startc);

      used[startc] = true;
      for(cv=0;cv<curves[startc]->cp[0].nv;cv++) {
	joined->cp[nc].vt[cv] = curves[startc]->cp[0].vt[cv];
      }
      
      for(thisc=order[startc];thisc != startc;thisc = order[thisc]) {
	
	for(i=0;i<curves[thisc]->cp[0].nv;i++) {
	  joined->cp[nc].vt[cv++] = curves[thisc]->cp[0].vt[i];
	}
	used[thisc] = true;
	printf("%d-",thisc);

      }
      
      printf("%d has %d vertices.\n",startc,joined->cp[nc].nv);
      nc++;
      
    }
    
  }

  plc_fix_wrap(joined);

  free(open);
  free(nv);
  free(used);
  
  /* Now output to a file */
   
  if (strlen(corefile->basename[0]) > sizeof(outfile_name)-20) {
    
    fprintf(stderr,"joinvect: Ridiculously long input filename "
	    "can't be parsed.\n");
    exit(1);
    
  }
  
  sprintf(outfile_name,"%s",corefile->basename[0]);
  
  if (strstr(outfile_name,".vect") != NULL) {
    
    sprintf(strstr(outfile_name,".vect"),".join.vect");
    
  } else {
    
    sprintf(outfile_tail,".join.vect");	 
    strcat(outfile_name,outfile_tail);
    
  }
  
  outfile_fptr = fopen(outfile_name,"w");
  
  if (outfile_fptr == NULL) {
    
    fprintf(stderr,
	    "joinvect: Couldn't open %s for writing.\n",outfile_name);
    exit(1);
    
  }
  
  /* Now write the tube to the file. */
  
  plc_write(outfile_fptr,joined);
  fclose(outfile_fptr);
  plc_free(joined);

  printf("joinvect: Wrote output file to %s.\n",outfile_name);
  
  /* Now do housekeeping */
  
  for(i=0;i<ncurves;i++) {
    
    plc_free(curves[i]);
    
  }
  
  exit(0);
  
}




