/*

  tsporder.c : Travelling salesman ordering for curves passed to joinvect.
               Obviously, this is only going to work if ncurves is small. 

*/

#include "joinvect.h"

#define dist(i,j) glob_distances[(i)*ncurves+(j)]

double *glob_distances = 0;

void tsp_recurser(int level,int ncurves,bool *used,
		  double thislen,double *minlen,
		  int order[],int workingorder[])

/* Searches for next links to add at this level. */

{

  int i;

  if (level == ncurves) { /* We've completed a circuit. Check if this wins. */

    if (thislen < *minlen) { 

      for(i=0;i<ncurves;i++) { order[i] = workingorder[i]; }
      *minlen = thislen;

    }

    return;

  }

  /* We are in the middle of building. Loop over the possibilities for this var. */

  for(i=0;i<ncurves;i++) {

    if (!used[i]) { /* if already used, ignore this option */

      if (thislen + dist(level,i) < *minlen) { /* if too long, ignore this option */

	used[i] = true;
	workingorder[level] = i;
	
	tsp_recurser(level+1,ncurves,used,thislen+dist(level,i),minlen,order,workingorder);
	
	used[i] = false;

      }

    }

  }

}

void plc_reverse(plCurve *L)

/* Reverses the orientation of L. */

{
  int cp,vt;
  plc_vector swap;

  for(cp=0;cp<L->nc;cp++) {

    for(vt=0;vt<L->cp[cp].nv/2;vt++) {

      swap = L->cp[cp].vt[vt];
      L->cp[cp].vt[vt] = L->cp[cp].vt[L->cp[cp].nv-vt-1];
      L->cp[cp].vt[L->cp[cp].nv-vt-1] = swap;

    }
  }
}

plc_vector plc_start(plCurve *L,int cp)

{
  return L->cp[cp].vt[0];
}

plc_vector plc_end(plCurve *L,int cp)
{
  return L->cp[cp].vt[L->cp[cp].nv-1];
}

void tsp_order(int ncurves,plCurve **curves,int *order)

/* Orders the curves so that every end vector is joined to a start vector. */
/* Note that we may have to reverse certain curves in order to get the best */
/* result. The resulting array is to be read as follows:

   2 4 1 0 3

   means 0 -> 2, 1 -> 4, 2 -> 1, 3 -> 0, and 4 -> 3, creating loop

   0 -> 2 -> 1 -> 4 -> 3 -> 0.

   Curves that are reversed are reversed in the plCurve array. */

{
  int i,j,k;
  bool   *used;

  glob_distances = calloc(ncurves*ncurves,sizeof(double));
  used = calloc(ncurves,sizeof(double));

  double minlen = 1e80;
  int    *reversed,*bestreversed,*workingorder;

  reversed = calloc(ncurves,sizeof(int));
  workingorder = calloc(ncurves,sizeof(int));
  bestreversed = calloc(ncurves,sizeof(int));

  for(k=0;k<pow(2,ncurves);k++) {

    /* We rebuild the distance array depending on the orientation of the curves. */

    for(i=0;i<ncurves;i++) {
      
      for(j=0;j<ncurves;j++) {
	
	dist(i,j) = plc_distance(
				 reversed[i] ? plc_start(curves[i],0) : plc_end(curves[i],0),
				 reversed[j] ? plc_end(curves[j],0) : plc_start(curves[j],0)
				 );
	
      }
      
    }

    for(i=0;i<ncurves;i++) { workingorder[i] = 0; }

    double old_minlen;
    old_minlen = minlen;

    tsp_recurser(0,ncurves,used,0.0,&minlen,order,workingorder);

    if (minlen < old_minlen) { /* We won! Quick, store the reversals... */

      for(i=0;i<ncurves;i++) { bestreversed[i] = reversed[i]; }

    }

    /* Now increment the reversed keys */

    for(j=0;j<ncurves;j++) {

      reversed[j]++;                      // increment this bit
      if (reversed[j] == 1) { break; }    // if no carry, we're done     
      reversed[j] = 0;                    // if carry, we set this bit to 0 and continue to the next one.
      
    }


  }

  /* Now we're done. Actually reverse the needed curves. */

  for(i=0;i<ncurves;i++) {

    if (bestreversed[i]) { plc_reverse(curves[i]); }

  }

  free(workingorder);
  free(used);
  free(glob_distances);
  glob_distances = NULL;

}
  


	
