#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif

#ifdef HAVE_TIME_H
#include <time.h>
#endif

#include "tube.h"
#include "plCurve.h"

 
plCurve *plCurve_equalize_sides(plCurve *L, int *target_verts)

  /* Procedure divides each polyline into target_verts sides, all
     of nearly equal length, by replacing the vertices of the ith pline
     with target_verts[i] vertices, spaced equally _along the curve_. 

     The procedure generates a new link with the target_vert counts.  
     We lose the color information from the original link in this version. */

{
  int             i,j,cmp;
  plc_vector  disp;
  double          side_length,length_so_far,t,seg_length;
  double         *component_lengths;

  plCurve   *new_L;
  int       *cc;
  bool      *open;

  assert(L != NULL);
  assert(target_verts != NULL);

  /* Allocate new memory for the revised link. */

  open = (bool *)(calloc(L->nc,sizeof(bool)));
  cc   = (int *)(calloc(L->nc,sizeof(int)));

  for(i=0;i<L->nc;i++) {

   open[i] = L->cp[i].open;
   cc[i]   = L->cp[i].cc;

  }

  new_L = plc_new(L->nc,target_verts,open,cc);
  free(open);
  free(cc);

  /* Now start computing the new link */

  component_lengths = (double *)(calloc(L->nc,sizeof(double)));
  plc_arclength(L,component_lengths);

  for(cmp=0;cmp<L->nc;cmp++) {

    side_length = component_lengths[cmp]/
      (double)(L->cp[cmp].open ? target_verts[cmp] - 1: target_verts[cmp]);

    new_L->cp[cmp].vt[0] = L->cp[cmp].vt[0];
    length_so_far = 0; t = 0; j = 1;

    /* We are now ready to loop through pline. */

    for(i=0;j<target_verts[cmp];i++) {
      
      disp = plc_vect_diff(L->cp[cmp].vt[i+1],L->cp[cmp].vt[i]);
      seg_length = plc_M_norm(disp);
      t = 0;
      
      while (t < 1) { /* While on the current segment of pline */
    
	if (side_length - length_so_far < (1-t)*seg_length) {
	  
	  /* If there is enough length here to end the new segment... */
	  
	  t += (side_length - length_so_far)/(seg_length);
	  plc_M_vlincomb(new_L->cp[cmp].vt[j],1,L->cp[cmp].vt[i],t,disp);
	  j++; length_so_far = 0;
	  
	} else {
	  
	  /* If not, we'll advance length_so_far, reset t, and increment i */

	  length_so_far += (1-t)*seg_length;
	  t = 1;
	  
	}
	
      }

    }
    
    /* Now, we should have generated all the new vertices. */
    /* If the polyline is.open, we need to set the last vert manually. */

    if (L->cp[cmp].open) {
    
      new_L->cp[cmp].vt[target_verts[cmp]-1] = L->cp[cmp].vt[L->cp[cmp].nv-1];

    }
    
  }

  plc_fix_wrap(new_L);
  return new_L;

}


plc_vector *random_framezeros(plCurve *link)

/* Procedure generates random initial vectors for orthonormal
   frames on the components of link. */

{
  plc_vector *fz_buf;
  plc_vector T;
  int i,j;
  double dprod;
  
  fz_buf = calloc(link->nc,sizeof(plc_vector));
  assert(fz_buf != NULL);

  srand((int)(time(NULL)));

  bool ok;
  
  for(i=0;i<link->nc;i++) {

    T = plc_mean_tangent(link,i,0,&ok);
    assert(ok);

    for(j=0;j<200;j++) {/* We worry about picking R and fz_buf[i] colinear. */

      fz_buf[i] = plc_random_vect();
      dprod = plc_M_dot(fz_buf[i],T);
      if (fabs(dprod) < 0.9) break;

    }
      
    assert(j < 200); /* And assume we've succeeded after 200 tries. */
    
    plc_M_scale_vect(dprod,T);		/* Ensure that dot(T,fz_buf[i]) = 0. */
    plc_M_sub_vect(fz_buf[i],T);
    fz_buf[i] = plc_normalize_vect(fz_buf[i],&ok);

    assert(ok);
  }

  return fz_buf;

}

void plCurve_bishop_frame(plCurve *link,
			       plc_vector *framezero,
			       plCurve **frameA, plCurve **frameB)

/* Procedure creates the "Bishop frame" on each component of the link
   with initial vectors given by "framezeros". */

{
  plc_vector T;
  int cp, vt;
  double dprod;

  /* First, we allocate space for the frame. */

  (*frameA) = plc_copy(link);
  (*frameB) = plc_copy(link);

  /* Now we are ready to fill it in. */

  bool ok;
			       
  for(cp=0;cp < link->nc;cp++) {

    T = plc_mean_tangent(link,cp,0,&ok);
    assert(ok);

    if (plc_M_dot(T,framezero[cp]) > 1e-6) {

      /* The initial vector is not normal to T,
         so we fix it. */

      dprod = plc_M_dot(T,framezero[cp]);
      plc_M_vmadd(framezero[cp],-dprod,T);
      framezero[cp] = plc_normalize_vect(framezero[cp],&ok);
      assert(ok);

    }

    (*frameA)->cp[cp].vt[0] = framezero[cp];
    (*frameB)->cp[cp].vt[0] = plc_cross_prod(T,framezero[cp]);
    (*frameB)->cp[cp].vt[0] = plc_normalize_vect((*frameB)->cp[cp].vt[0],&ok);
    assert(ok);
    
    for(vt=1;vt < link->cp[cp].nv;vt++) {

      T = plc_mean_tangent(link,cp,vt,&ok);
      assert(ok);

      (*frameA)->cp[cp].vt[vt] = (*frameA)->cp[cp].vt[vt-1];

      dprod = plc_M_dot(T,(*frameA)->cp[cp].vt[vt]);
      plc_M_vmadd((*frameA)->cp[cp].vt[vt],-dprod,T);
      (*frameA)->cp[cp].vt[vt] = plc_normalize_vect((*frameA)->cp[cp].vt[vt],
						    &ok);
      assert(ok);

      (*frameB)->cp[cp].vt[vt] = 
	plc_normalize_vect(plc_cross_prod(T,(*frameA)->cp[cp].vt[vt]),&ok);
      assert(ok);

    }
      
  }

  /* We've now generated the frame. */

}

void plc_force_closed( plCurve *inLink )

     /* Forces each open component of inLink to be closed. */

{
  int cItr, fixItr;
  plc_vector error;
  double efrac;

  for(cItr=0;cItr<inLink->nc;cItr++) {

    if (inLink->cp[cItr].open) { 

      error = plc_vect_diff(inLink->cp[cItr].vt[inLink->cp[cItr].nv-1],
			    inLink->cp[cItr].vt[0]);

      for( fixItr=0; fixItr < inLink->cp[cItr].nv; fixItr++) {

	efrac = (double)fixItr/(double)inLink->cp[cItr].nv-1;

	inLink->cp[cItr].vt[fixItr] = plc_vmadd(inLink->cp[cItr].vt[fixItr],
						efrac,error);

      }

      /* At this point, the last and first vertices are the same. */

      inLink->cp[cItr].nv--;
      inLink->cp[cItr].open = false;
      
    }

  }

  plc_fix_wrap(inLink);

}
						
      
      
      

      

      
   
